#! /usr/bin/env python3

import hashlib
import operator
import os
import sys
import time

READ_CHUNKS = 64 * 1024


def _printable_time(tstamp):
    """Return a nice printable time."""
    text = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(tstamp))
    text += ",%03d" % ((tstamp % 1) * 1000,)
    return text


def main(alldirnames):
    """Main entry point."""
    allnodes = []
    for dirname in alldirnames:
        print("Analyzing", repr(dirname))
        c = 0
        for dirpath, dirnames, filenames in os.walk(dirname):
            for fname in filenames:
                fullpath = os.path.join(dirpath, fname)
                stat = os.stat(fullpath)
                allnodes.append((fullpath, fname, stat.st_size, stat.st_mtime))
                c += 1
                if not c % 1000:
                    print("\r    ", c, end='')
                    sys.stdout.flush()
        print("\r           ")
    print("\rFound %d nodes" % len(allnodes))

    print("Reducing information")
    redux_size = {}
    redux_name = {}
    for fullpath, fname, size, mtime in allnodes:
        redux_size.setdefault(size, []).append((fullpath, mtime))
        redux_name.setdefault(fname, []).append((fullpath, mtime, size))

    print("\nNodes with same name:")
    items = sorted(redux_name.items(), key=lambda it: sorted(it[1])[0][0])
    for fname, nodesinfo in items:
        if len(nodesinfo) > 1:
            print()
            for fullpath, mtime, size in nodesinfo:
                nice_time = _printable_time(mtime)
                print("  %10d  %s  %r" % (size, nice_time, fullpath))

    print("\nNodes with same content:")
    items = sorted(redux_size.items(), key=lambda it: sorted(it[1])[0][0])
    for size, nodesinfo in items:
        if len(nodesinfo) > 1:
            # do another reduction, per hash of the file
            subredux = {}
            for fullpath, mtime in nodesinfo:
                hasher = hashlib.sha1()
                with open(fullpath, 'rb') as fh:
                    while True:
                        data = fh.read(READ_CHUNKS)
                        if not data:
                            break
                        hasher.update(data)
                digest = hasher.digest()
                subredux.setdefault(digest, []).append((fullpath, mtime))

            # process the subreduction
            for hashdig, nodesinfo in subredux.items():
                if len(nodesinfo) > 1:
                    print()
                    for fullpath, mtime in sorted(nodesinfo, key=operator.itemgetter(1)):
                        nice_time = _printable_time(mtime)
                        print("  %10d  %s  %r" % (size, nice_time, fullpath))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: dupes.py dir1 [dir2 [...]]")
        exit()

    main(sys.argv[1:])
